//! Differential Fault Analysis (DFA) attack on DES.

use crate::des::{cut_in_halves, des_encrypt, permute, E, IP, SBOX};
use std::collections::HashMap;

/// Inverse of P-permutation (32 -> 32)
const P_INV: [u8; 32] = [
    9, 17, 23, 31, 13, 28, 2, 18, 24, 16, 30, 6, 26, 20, 10, 1, 8, 14, 25, 3, 4, 29, 11, 19, 32,
    12, 22, 7, 5, 27, 15, 21,
];

/// Inverse PC1 (64 -> 56)
const PC1_INV: [u8; 64] = [
    8, 16, 24, 56, 52, 44, 36, 0, 7, 15, 23, 55, 51, 43, 35, 0, 6, 14, 22, 54, 50, 42, 34, 0, 5,
    13, 21, 53, 49, 41, 33, 0, 4, 12, 20, 28, 48, 40, 32, 0, 3, 11, 19, 27, 47, 39, 31, 0, 2, 10,
    18, 26, 46, 38, 30, 0, 1, 9, 17, 25, 45, 37, 29, 0,
];

/// Inverse PC2 (56 -> 48)
const PC2_INV: [u8; 56] = [
    5, 24, 7, 16, 6, 10, 20, 18, 0, 12, 3, 15, 23, 1, 9, 19, 2, 0, 14, 22, 11, 0, 13, 4, 0, 17, 21,
    8, 47, 31, 27, 48, 35, 41, 0, 46, 28, 0, 39, 32, 25, 44, 0, 37, 34, 43, 29, 36, 38, 45, 33, 26,
    42, 0, 30, 40,
];

/// Determine which S‐boxes were affected by each fault
fn select_faults(correct: u64, faults: &[u64]) -> HashMap<usize, Vec<usize>> {
    let mut map: HashMap<usize, Vec<usize>> = (0..8).map(|i| (i, Vec::new())).collect();
    for (i, &f) in faults.iter().enumerate() {
        let diff = correct ^ f;
        let perm = permute(diff, &IP, 64);
        let (_l16, r16) = cut_in_halves(perm);
        let expanded = permute(r16 as u64, &E, 32);
        // split into 8 chunks of 6 bits
        for s in 0..8 {
            let block = ((expanded >> (42 - 6 * s)) & 0x3F) as u8;
            if block != 0 {
                map.get_mut(&s).unwrap().push(i);
            }
        }
    }
    map
}

/// returns the single common element of the intersection of multiple Vec<u8>
fn common(lists: &[Vec<u8>]) -> u8 {
    let mut set: std::collections::HashSet<u8> = lists[0].iter().copied().collect();
    for lst in &lists[1..] {
        set = set
            .intersection(&lst.iter().copied().collect())
            .copied()
            .collect();
    }
    assert!(set.len() == 1, "expected exactly one common solution");
    *set.iter().next().unwrap()
}

/// 48-bit mask to isolate the 6 bits for s‐box `s`
fn mask48(s: usize) -> u64 {
    let start = 6 * (7 - s);
    let mask: u64 = 0x3Fu64 << start;
    mask
}

/// 32-bit mask to isolate the 4-bit output of s‐box `s`
fn mask32(s: usize) -> u32 {
    let start = 4 * (7 - s);
    0xFu32 << start
}

/// Given expanded R15 and a candidate subkey `k6`, compute (row, col)
fn calc_line_col(s: usize, expanded: u64, k6: u8) -> (usize, usize) {
    let mut six = ((expanded & mask48(s)) >> (6 * (7 - s))) as u8;
    six ^= k6;
    let row = (((six & 0x20) >> 4) | (six & 0x01)) as usize;
    let col = ((six & 0x1E) >> 1) as usize;
    (row, col)
}

/// Check if candidate k6 is valid for this fault in s‐box s
fn solution_valid(
    s: usize,
    p1_diff: u32,
    row: usize,
    col: usize,
    row_f: usize,
    col_f: usize,
) -> bool {
    let out_diff = (p1_diff & mask32(s)) >> (4 * (7 - s));
    let x = SBOX[s][row][col] ^ SBOX[s][row_f][col_f];
    out_diff == x as u32
}

/// Find the 48-bit round‐16 subkey K16
pub fn find_k16(correct: u64, faults: &[u64]) -> u64 {
    let sboxes = select_faults(correct, faults);
    // get L16 and R15 from correct ciphertext
    let perm = permute(correct, &IP, 64);
    let (l16, r15) = cut_in_halves(perm);

    let mut k16: u64 = 0;
    for s in 0..8 {
        let idxs = &sboxes[&s];
        // for each fault index for this sbox, collect the faulty ciphertext
        let mut sol_lists: Vec<Vec<u8>> = Vec::new();
        let expanded = permute(r15 as u64, &E, 32);
        for &i in idxs.iter().take(6) {
            let perm_f = permute(faults[i], &IP, 64);
            let (l16f, r15f) = cut_in_halves(perm_f);
            let p1_diff = permute((l16 ^ l16f) as u64, &P_INV, 32) as u32;
            let expf = permute(r15f as u64, &E, 32);
            // try all 64 candidates
            let mut cands = Vec::new();
            for k6 in 0u8..64u8 {
                let (r0, c0) = calc_line_col(s, expanded, k6);
                let (rf, cf) = calc_line_col(s, expf, k6);
                if solution_valid(s, p1_diff, r0, c0, rf, cf) {
                    cands.push(k6);
                }
            }
            sol_lists.push(cands);
        }
        let sol = common(&sol_lists) as u64;
        k16 = (k16 << 6) | sol;
    }
    k16
}

/// Brute‐force the 8 missing bits to recover the 64‐bit master key (parity bits are set to 0 for padding)
pub fn find_k64(clear: u64, cipher: u64, k16: u64) -> u64 {
    // invert PC2 then PC1
    let k56 = permute(permute(k16, &PC2_INV, 48), &PC1_INV, 56);
    // 8 unknown bits
    let positions = [50, 49, 45, 44, 13, 10, 6, 4];
    for mask in 0u16..(1 << 8) {
        let mut candidate = k56;
        for (i, &position) in positions.iter().enumerate() {
            if (mask & (1 << i)) != 0 {
                candidate |= 1u64 << position;
            }
        }
        if des_encrypt(clear, candidate) == cipher {
            return candidate;
        }
    }
    panic!("Failed to recover full key");
}

/// Set odd parity on each byte
pub fn apply_parity(mut k: u64) -> u64 {
    for byte in 0..8 {
        let mut count = 0;
        for b in 1..8 {
            if (k & (1u64 << (byte * 8 + b))) != 0 {
                count += 1;
            }
        }
        if count % 2 == 0 {
            k |= 1u64 << (byte * 8);
        } else {
            k &= !(1u64 << (byte * 8));
        }
    }
    k
}
